#!/usr/bin/env python3

'''
logging_better module;
'''

import logging
import types

DEBUG   = logging.DEBUG
VERBOSE = logging.INFO - 1
INFO    = logging.INFO
WARN    = logging.WARN
ERROR   = logging.ERROR

def d(msg, *args, **kwargs):

    '''
    log a message at debug level;
    '''

    return logging.log(DEBUG, msg, *args, **kwargs)

def v(msg, *args, **kwargs):

    '''
    log a message at verbose level;
    '''

    return logging.log(verbose, msg, *args, **kwargs)

def i(msg, *args, **kwargs):

    '''
    log a message at info level;
    '''

    return logging.log(info, msg, *args, **kwargs)

def w(msg, *args, **kwargs):

    '''
    log a message at warn level;
    '''

    return logging.log(warn, msg, *args, **kwargs)

def e(msg, *args, **kwargs):

    '''
    log a message at error level;
    '''

    return logging.log(error, msg, *args, **kwargs)

def getLogger(name=None):

    '''
    wrapper of `logging.getLogger` with additional methods bound to the returned
    logger;
    '''

    logger = logging.getLogger(name=name)

    if not hasattr(logger, 'd'):
        def d(self, msg, *args, **kwargs):
            return self.log(DEBUG, msg, *args, **kwargs)
        logger.d = types.MethodType(d, logger)

    if not hasattr(logger, 'v'):
        def v(self, msg, *args, **kwargs):
            return self.log(VERBOSE, msg, *args, **kwargs)
        logger.v = types.MethodType(v, logger)

    if not hasattr(logger, 'i'):
        def i(self, msg, *args, **kwargs):
            return self.log(INFO, msg, *args, **kwargs)
        logger.i = types.MethodType(i, logger)

    if not hasattr(logger, 'w'):
        def w(self, msg, *args, **kwargs):
            return self.log(WARN, msg, *args, **kwargs)
        logger.w = types.MethodType(w, logger)

    if not hasattr(logger, 'e'):
        def e(self, msg, *args, **kwargs):
            return self.log(ERROR, msg, *args, **kwargs)
        logger.e = types.MethodType(e, logger)

    return logger




#!/usr/bin/env python3

'''utility module;'''

import sys

def die(s):

    '''
    print an error message and exit;
    '''

    print('error: ' + s, file=sys.stderr)
    sys.exit(1)

def printout(s):
    '''
    Print to stdout and flush.
    '''
    print(s, file=sys.stdout, flush=True)

def printerr(s):
    '''
    Print to stderr and flush.
    '''
    print(s, file=sys.stderr, flush=True)

def is_window_in_viewport(window, viewport):
    '''
    Check whether window is in viewport.

    Parameters
    ----------
    window : object
        A window object.
    viewport : list
        A viewport.

    Returns
    -------
    bool
        True iff window is in viewport.
    '''
    # Get window geometry.
    win_geo = window.get_geometry()
    # Get root window.
    root = win_geo.root
    # Translate coords (window -> root).
    win_in_root_geo = root.translate_coords(window, win_geo.x, win_geo.y)

    wl = win_in_root_geo.x
    wt = win_in_root_geo.y
    wr = win_in_root_geo.x + win_geo.width - 1
    wb = win_in_root_geo.y + win_geo.height - 1

    # Get root geometry.
    root_geo = root.get_geometry()

    vl = viewport[0]
    vt = viewport[1]
    vr = viewport[0] + root_geo.width - 1
    vb = viewport[1] + root_geo.height - 1

    return not (wl > vr or wt > vb or wr < vl or wb < vt)




#!/usr/bin/env python3

'''
Layout module.

This module contains layout handlers which lay windows into different layout
patterns.

To implement a new layout, define a new layout handler and add it to
`LAYOUT_HANDLERZ`.
'''

from awd.api import get_workarea
from awd.api import move_resize_window
from awd.api import set_active_window
import awd.logging_better as log

def _order_windows(windows, active_window):
    '''
    Order windows. Active window comes last.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    list
        A list of window objects, in which active window comes last.
    '''
    tmp = None
    ans = []
    for window in windows:
        if window.id == active_window.id:
            tmp = window
        else:
            ans.append(window)
    if tmp is not None:
        ans.append(tmp)
    return ans

def _move_resize_window(window, x, y, w, h):
    '''
    Changes the size and location of the specified window.

    This function takes border width into account.

    Parameters
    ----------
    window
        A window object.
    x
        X coord.
    y
        Y coord.
    w
        Width.
    h
        Height.

    Returns
    -------
    None
        None.
    '''
    # Find top window.
    top = window
    while True:
        resp = top.query_tree()
        if top == resp.root or resp.parent == resp.root:
            break
        top = resp.parent

    # Get left and top offsets. This is done by translating coord (0, 0) in the
    # given window to the top window.
    resp = top.translate_coords(window, 0, 0)
    l = resp.x
    t = resp.y

    # Get right and bottom offsets. This is done by computing size difference.
    win_geo = window.get_geometry()
    top_geo = top.get_geometry()
    r = top_geo.width - win_geo.width - l
    b = top_geo.height - win_geo.height - t

    # Move and resize window. WM will take care of the location, so we just need
    # to take care of the size.
    move_resize_window(window, x, y, w - l - r, h - t - b)
    log.d(['move_resize_window', x, y, w - l - r, h - t - b])

def cascade(windows, active_window):
    '''
    Cascade windows.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Get work area.
    workarea = get_workarea()

    ngrid = 16  # Magic: Number of grids on each axis.
    nwin = 10   # Magic: Number of grids a window occupies on each axis.

    xstep, ystep = workarea[2] // ngrid, workarea[3] // ngrid

    x, y, w, h = xstep, ystep, nwin * xstep, nwin * ystep
    for window in _order_windows(windows, active_window):
        log.d([window, x, y, w, h])
        _move_resize_window(window, x, y, w, h)
        set_active_window(window)
        if len(windows) == 1:
            break
        x += (ngrid - 2 - nwin) * xstep // (len(windows) - 1)
        y += (ngrid - 2 - nwin) * ystep // (len(windows) - 1)

def hstack(windows, active_window):
    '''
    Stack windows horizontally.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Get work area.
    workarea = get_workarea()

    x = workarea[0]
    y = workarea[1]
    w = workarea[2] // len(windows)
    h = workarea[3]

    for window in _order_windows(windows, active_window):
        log.d([window, x, y, w, h])
        _move_resize_window(window, x, y, w, h)
        x += w

def vstack(windows, active_window):
    '''
    Stack windows vertically.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Get work area.
    workarea = get_workarea()

    x = workarea[0]
    y = workarea[1]
    w = workarea[2]
    h = workarea[3] // len(windows)

    for window in _order_windows(windows, active_window):
        log.d([window, x, y, w, h])
        _move_resize_window(window, x, y, w, h)
        y += h

def tmain(windows, active_window):
    '''
    Two-row layout, top row is main row.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Process the active window.
    workarea = get_workarea()
    width = workarea[2]
    height = workarea[3] // 2
    _move_resize_window(active_window, 0, 0, width, height)

    # Process the rest.
    if len(windows) > 1:
        x, y, w, h = 0, height, width // (len(windows) - 1), height
        for i, window in enumerate(_order_windows(windows, active_window)):
            if i == len(windows) - 1:
                break
            _move_resize_window(window, x, y, w, h)
            x += w

def bmain(windows, active_window):
    '''
    Two-row layout, bottom row is main row.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Process the active window.
    workarea = get_workarea()
    width = workarea[2]
    height = workarea[3] // 2
    _move_resize_window(active_window, 0, height, width, height)

    # Process the rest.
    if len(windows) > 1:
        x, y, w, h = 0, 0, width // (len(windows) - 1), height
        for i, window in enumerate(_order_windows(windows, active_window)):
            if i == len(windows) - 1:
                break
            _move_resize_window(window, x, y, w, h)
            x += w

def lmain(windows, active_window):
    '''
    Two-col layout, left col is main col.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Process the active window.
    workarea = get_workarea()
    width = workarea[2] // 2
    height = workarea[3]
    _move_resize_window(active_window, 0, 0, width, height)

    # Process the rest.
    if len(windows) > 1:
        x, y, w, h = width, 0, width, height // (len(windows) - 1)
        for i, window in enumerate(_order_windows(windows, active_window)):
            if i == len(windows) - 1:
                break
            _move_resize_window(window, x, y, w, h)
            y += h

def rmain(windows, active_window):
    '''
    Two-col layout, right col is main col.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    # Process the active window.
    workarea = get_workarea()
    width = workarea[2] // 2
    height = workarea[3]
    _move_resize_window(active_window, width, 0, width, height)

    # Process the rest.
    if len(windows) > 1:
        x, y, w, h = 0, 0, width, height // (len(windows) - 1)
        for i, window in enumerate(_order_windows(windows, active_window)):
            if i == len(windows) - 1:
                break
            _move_resize_window(window, x, y, w, h)
            y += h

def _rowgrid(windows, active_window, nrows, ncols):
    '''
    Grid layout, row-major.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.
    nrows:
        Number of rows.
    ncols:
        Number of cols.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    if nrows * ncols < len(windows):
        log.w('Warning: Not enough space. Windows may overlap.')

    workarea = get_workarea()
    width = workarea[2] // ncols
    height = workarea[3] // nrows
    x, y, w, h = 0, 0, width, height
    for i, window in enumerate(_order_windows(windows, active_window)):
        y, x = divmod(i % (nrows * ncols), ncols)
        x *= width
        y *= height
        _move_resize_window(window, x, y, w, h)

def rowgrid22(desktop, active_window):
    return _rowgrid(desktop, active_window, 2, 2)
def rowgrid23(desktop, active_window):
    return _rowgrid(desktop, active_window, 2, 3)
def rowgrid24(desktop, active_window):
    return _rowgrid(desktop, active_window, 2, 4)
def rowgrid32(desktop, active_window):
    return _rowgrid(desktop, active_window, 3, 2)
def rowgrid33(desktop, active_window):
    return _rowgrid(desktop, active_window, 3, 3)
def rowgrid34(desktop, active_window):
    return _rowgrid(desktop, active_window, 3, 4)
def rowgrid42(desktop, active_window):
    return _rowgrid(desktop, active_window, 4, 2)
def rowgrid43(desktop, active_window):
    return _rowgrid(desktop, active_window, 4, 3)
def rowgrid44(desktop, active_window):
    return _rowgrid(desktop, active_window, 4, 4)

def _colgrid(windows, active_window, nrows, ncols):
    '''
    Grid layout, col-major.

    Parameters
    ----------
    windows : list
        A list of window objects.
    active_window : object
        The active window object.
    nrows:
        Number of rows.
    ncols:
        Number of cols.

    Returns
    -------
    None
        None.
    '''
    if len(windows) == 0:
        return

    if nrows * ncols < len(windows):
        log.w('Warning: Not enough space. Windows may overlap.')

    workarea = get_workarea()
    width = workarea[2] // ncols
    height = workarea[3] // nrows
    x, y, w, h = 0, 0, width, height
    for i, window in enumerate(_order_windows(windows, active_window)):
        x, y = divmod(i % (nrows * ncols), nrows)
        x *= width
        y *= height
        _move_resize_window(window, x, y, w, h)

def colgrid22(desktop, active_window):
    return _colgrid(desktop, active_window, 2, 2)
def colgrid23(desktop, active_window):
    return _colgrid(desktop, active_window, 2, 3)
def colgrid24(desktop, active_window):
    return _colgrid(desktop, active_window, 2, 4)
def colgrid32(desktop, active_window):
    return _colgrid(desktop, active_window, 3, 2)
def colgrid33(desktop, active_window):
    return _colgrid(desktop, active_window, 3, 3)
def colgrid34(desktop, active_window):
    return _colgrid(desktop, active_window, 3, 4)
def colgrid42(desktop, active_window):
    return _colgrid(desktop, active_window, 4, 2)
def colgrid43(desktop, active_window):
    return _colgrid(desktop, active_window, 4, 3)
def colgrid44(desktop, active_window):
    return _colgrid(desktop, active_window, 4, 4)

# Layout handlers.
LAYOUT_HANDLERZ = {
    'cascade': cascade,
    'hstack': hstack,
    'vstack': vstack,
    'bmain': bmain,
    'lmain': lmain,
    'rmain': rmain,
    'tmain': tmain,
    'rowgrid22': rowgrid22,
    'rowgrid23': rowgrid23,
    'rowgrid24': rowgrid24,
    'rowgrid32': rowgrid32,
    'rowgrid33': rowgrid33,
    'rowgrid34': rowgrid34,
    'rowgrid42': rowgrid42,
    'rowgrid43': rowgrid43,
    'rowgrid44': rowgrid44,
    'colgrid22': colgrid22,
    'colgrid23': colgrid23,
    'colgrid24': colgrid24,
    'colgrid32': colgrid32,
    'colgrid33': colgrid33,
    'colgrid34': colgrid34,
    'colgrid42': colgrid42,
    'colgrid43': colgrid43,
    'colgrid44': colgrid44,
}



#!/usr/bin/env python3

'''Xlib/EWMH API module.'''

from Xlib.display import Display
from ewmh import EWMH

# Display object.
disp = None

# EWMH object.
ewmh = None

def _get_display():
    '''
    Get the display object.

    Returns
    -------
    object
        The display object.
    '''
    global disp
    if disp is None:
        disp = Display()
    return disp

def _get_ewmh():
    '''
    Get the EWMH object.

    Returns
    -------
    object
        The EWMH object.
    '''
    global ewmh
    if ewmh is None:
        ewmh = EWMH()
    return ewmh

def get_windows():
    '''
    Get all X Windows managed by the Window Manager.

    Sticky windows are excluded.

    Returns
    -------
    list
        A list of windows managed by the Window Manager.
    '''
    ewmh = _get_ewmh()
    windows = []
    for w in ewmh.getClientList():
        wm_state = ewmh.getWmState(w, True)
        if '_NET_WM_STATE_STICKY' not in wm_state:
            windows.append(w)
    return windows

def get_desktop_geometry():
    '''
    Get the common size of all desktops.

    Returns
    -------
    list
        The common size of all desktops.
    '''
    ewmh = _get_ewmh()
    return list(ewmh.getDesktopGeometry())

def get_desktop_viewport():
    '''
    Get the top left corner of each desktop's viewport.

    Returns
    -------
    list
        The [top, left] corner of each desktop's viewport.
    '''
    ewmh = _get_ewmh()
    return list(ewmh.getDesktopViewPort())

def get_current_desktop():
    '''
    Get the index of the current desktop.

    Returns
    -------
    int
        The index of the current desktop.
    '''
    ewmh = _get_ewmh()
    return ewmh.getCurrentDesktop()

def get_active_window():
    '''
    Get the active window.

    Returns
    -------
    object
        The active window object.
    '''
    ewmh = _get_ewmh()
    return ewmh.getActiveWindow()

def get_workarea():
    '''
    Get the work area.

    Returns
    -------
    list
        The work area [x, y, w, h].
    '''
    ewmh = _get_ewmh()
    return list(ewmh.getWorkArea())

def get_wm_name(window):
    '''
    Get name of the window.

    Parameters
    ----------
    window : object
        A window object.

    Returns
    -------
    str
        Name of the window.
    '''
    ewmh = _get_ewmh()
    return ewmh.getWmName(window).decode()

def get_wm_desktop(window):
    '''
    Get the desktop the window is in.

    Parameters
    ----------
    window : object
        A window object.

    Returns
    -------
    int
        A desktop number.
    '''
    ewmh = _get_ewmh()
    return ewmh.getWmDesktop(window)

def set_active_window(window):
    '''
    Set the active window.

    Parameters
    ----------
    window : object
        A window object.

    Returns
    -------
    None
        None.
    '''
    ewmh = _get_ewmh()
    ewmh.setActiveWindow(window)
    ewmh.display.flush()

def move_resize_window(window, x, y, w, h):
    '''
    Changes the size and location of the specified window.

    Parameters
    ----------
    window
        A window object.
    x
        X coord.
    y
        Y coord.
    w
        Width.
    h
        Height.

    Returns
    -------
    None
        None.
    '''
    disp = _get_display()
    window.configure(x=x, y=y, width=w, height=h)
    disp.flush()




#!/usr/bin/env python3

'''
A wacky manager of window manager.
'''

import argparse
import argparse_better
import sys

from traceback import print_exc

from awd.api import get_active_window
from awd.api import get_current_desktop
from awd.api import get_desktop_viewport
from awd.api import get_windows
from awd.api import get_wm_desktop
from awd.api import get_wm_name
from awd.api import set_active_window
from awd.util import is_window_in_viewport
from awd.util import printerr
from awd.util import die

##  program name;
prog = 'awd'

epilog = '''\
available layouts:
    rowgrid[x][y]       row-major grid layout (2 <= x, y <= 4)
    colgrid[x][y]       col-major grid layout (2 <= x, y <= 4)
'''

def filter_windows(windows, desktop, viewport, excludes):
    '''
    Filter out windows which are not in the desktop or not in the viewport.

    Parameters
    ----------
    windows : list
        A list of windows.
    desktop : int
        A desktop number.
    viewport : list
        A viewport.
    excludes : list
        A list of window name exclude patterns.

    Returns
    -------
    list
        A list of windows that pass the filter.
    '''
    ans = []
    for window in windows:

        # Filter by desktop.
        if get_wm_desktop(window) != desktop:
            continue

        # Filter by viewport.
        if not is_window_in_viewport(window, viewport):
            continue

        # Filter by window name.
        wm_name = get_wm_name(window)
        if any([x in wm_name for x in excludes]):
            continue

        ans.append(window)
    return ans

def layout_windows(windows, layout, active_window):
    '''
    Layout windows.

    Parameters
    ----------
    windows : list
        A list of windows.
    layout : str
        Layout name.
    active_window : object
        The active window object.
    '''
    LAYOUT_HANDLERZ[layout](windows, active_window)

def parse_args():

    '''
    parse command line arguments;
    '''

    parser = argparse.ArgumentParser(
        prog=prog,
        description='a window director;',
        formatter_class=argparse_better.HelpFormatter,
        add_help=False,
    )

    parser.add_argument(
        '-h', '--help',
        action='help',
        help='display help message;',
    )
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help='debug mode;',
    )
    parser.add_argument(
        '-e', '--exclude',
        action='append',
        default=[],
        metavar='pattern',
        help='excluded window names;',
    )

    layouts = parser.add_argument_group('layouts')

    layouts.add_argument(
        '--cascade',
        action='store_true',
        help='cascade;',
    )
    layouts.add_argument(
        '--horizontal',
        action='store_true',
        help='tile horizontally;',
    )
    layouts.add_argument(
        '--vertical',
        action='store_true',
        help='tile vertically;',
    )
    layouts.add_argument(
        '--left',
        action='store_true',
        help='main window on the left;',
    )
    layouts.add_argument(
        '--right',
        action='store_true',
        help='main window on the right;',
    )
    layouts.add_argument(
        '--top',
        action='store_true',
        help='main window on the top;',
    )
    layouts.add_argument(
        '--bottom',
        action='store_true',
        help='main window on the bottom;',
    )
    layouts.add_argument(
        '--grid',
        action='store_true',
        help='grid;',
    )

    layout_opts = parser.add_argument_group('layout options')

    layout_opts.add_argument(
        '--row',
        type=int,
        metavar='row',
        help='number of grid rows;',
    )
    layout_opts.add_argument(
        '--col',
        type=int,
        metavar='col',
        help='number of grid cols;',
    )

    args = parser.parse_args()

    ##  check conflicts;
    mutex = [ k for k in [
        'cascade', 'horizontal', 'vertical', 'left', 'right', 'top', 'bottom',
        'grid',
    ] if getattr(args, k, False) ]
    if len(mutex) > 1:
        die('layout confict: ' + ', '.join(mutex))

    return args

def main():

    '''
    main function;
    '''

    ##  parse args;
    args = parse_args()
    print(args)
    sys.exit()

    layout = args.layout
    exclude = args.exclude

    # Set log level.
    logger.setLevel({v: k for k, v in LOGLEVELZ.items()}[loglevel])

    # Get all windows managed by WM.
    windows = get_windows()
    logger.d('all windows %s', windows)

    # Get current active window.
    active_window = get_active_window()
    logger.d('active_window %s', active_window)

    # Get current desktop.
    desktop = get_current_desktop()
    logger.d('current desktop %s', desktop)

    # Get current viewport.
    viewport = get_desktop_viewport()
    logger.d('current viewport %s', viewport)

    # Filter windows by desktop and viewport.
    windows = filter_windows(windows, desktop, viewport, exclude)
    logger.d('filtered windows %s', windows)

    # Layout windows.
    layout_windows(windows, layout, active_window)

    # Activate the original active window.
    set_active_window(active_window)

if __name__ == '__main__':
    main()

